unit vvvOCR;

{$mode objfpc}{$H+}

interface

uses
 Classes, SysUtils;

type
 TgrayImageData = record
  Width, Height: word;
  Data: TBytes;
 end;

 PgrayImageData = ^TgrayImageData;

 TgrayImage = class;

 TfinderCallBackProc = procedure(const Sender: TgrayImage; const r: TRect; var stop: boolean) of object;

 TgrayImage = class
 private
  fOneBit: boolean;
  fData: TgrayImageData;
  procedure setWidth(w: word);
  procedure setHeight(h: word);
  function getPixel(x, y: word): byte;
  procedure setPixel(x, y: word; v: byte);
  function getScanline(y: word): PByteArray;
 protected
  procedure DataChanged; virtual;
  property Data: TBytes read fData.Data;
  property rawData: TgrayImageData read fData;
 public
  constructor Create; overload;
  constructor Create(const w, h: word; const dt: PByte); overload;
  destructor Destroy; override;
  procedure LoadFromStream(s: TStream);
  procedure SaveToStream(s: TStream);
  procedure FindObjects(cbp: TfinderCallBackProc; const rect: TRect; const invert: boolean = False);
  procedure CopyRect(r: TRect; src: TgrayImage);
  procedure Copy(src: TgrayImage); overload;
  procedure Copy(src: TgrayImageData); overload;
  procedure SetData(const w, h: word; const dt: PByte);
  procedure Normalize; virtual;
  procedure Convert1Bit(const invert: boolean = False);
  procedure Normalize1Bit(const sampleLength: word = 20);
 public
  property Pixel[x, y: word]: byte read getPixel write setPixel;
  property ScanLine[y: word]: PByteArray read getScanline;
  property Width: word read fData.width write setWidth;
  property Height: word read fData.height write setHeight;
  property OneBit: boolean read fOneBit;
 end;

 TGrayImageClass = class of TgrayImage;

 TabImageData = record
  symbol: ansiChar;
  maxWeight: word;
  image: TgrayImageData;
 end;

 PabImageData = ^TabImageData;

 TabImageDataArray = array of TabImageData;

 TabImages = class
 private
  fHeight: word;
  fAngle: integer;
  fRatio: byte;
  function getCount: integer;
 public
  Images: TabImageDataArray;
 public
  constructor Create(aHeight: word; aAngle: integer; aRatio: byte);
  destructor Destroy; override;
  procedure LoadFromStream(s: TStream);
  procedure SaveToStream(s: TStream);
 public
  property Height: word read fHeight;
  property Angle: integer read fAngle;
  property Count: integer read getCount;
  property Ratio: byte read fRatio;
 end;

 PabImages = ^TabImages;
 TabImagesArray = array of TabImages;

 TbasicOCR = class
 private
  fCharWeight: byte;
 protected
  fText: ansiString;
  Images: TabImagesArray;
  function SearchChar(img: TgrayImage; r: TRect; h: byte; a: integer; p: byte; valid: ansiString): ansiChar;
  property CharWeight: byte read fCharWeight;
 public
  constructor Create;
  destructor Destroy; override;
  procedure ClearSymbols;
  procedure LoadFromStream(s: TStream);
  procedure LoadFromFile(const fn: string);
  function Search(const img: TgrayImage; const r: TRect): boolean; virtual; overload;
  function Search(const img: TgrayImage): boolean; virtual; overload;
 public
  minWeight: byte;
  property Text: ansiString read fText;
 end;

const
 ocrWeights: array [0..1, 0..4] of SmallInt = ((5, 2, 0, -2, -5),(-5, -2, 0, 2, 5));

procedure LoadGrayImage(const s: TStream; var img: TgrayImageData);
procedure SaveGrayImage(const s: TStream; const img: TgrayImageData);

implementation

procedure LoadGrayImage(const s: TStream; var img: TgrayImageData);
begin
 s.Read(img.Width, SizeOf(img.Width));
 s.Read(img.Height, SizeOf(img.Height));
 setLength(img.Data, img.Width * img.Height);
 s.Read(img.Data[0], img.Width * img.Height);
end;

procedure SaveGrayImage(const s: TStream; const img: TgrayImageData);
begin
 s.Write(img.Width, SizeOf(img.Width));
 s.Write(img.Height, SizeOf(img.Height));
 s.Write(img.Data[0], img.Width * img.Height);
end;

//################################################################## TgrayImage

constructor TgrayImage.Create;
begin
 fData.Width:= 0;
 fData.Height:= 0;
 SetLength(fData.Data, 0);
 fOneBit:= False;
end;

constructor TgrayImage.Create(const w, h: word; const dt: PByte);
begin
 Create;
 SetData(w, h, dt);
end;

destructor TgrayImage.Destroy;
begin
 setLength(fData.Data, 0);
 inherited;
end;

procedure TgrayImage.setWidth(w: word);
begin
 fData.Width:= w;
 setLength(fData.Data, fData.Width * fData.Height);
 DataChanged;
end;

procedure TgrayImage.setHeight(h: word);
begin
 fData.Height:= h;
 setLength(fData.Data, fData.Width * fData.Height);
 DataChanged;
end;

function TgrayImage.getPixel(x, y: word): byte;
begin
 if (x >= fData.Width) or (y >= fData.Height)
 then Result:= 1
 else Result:= fData.Data[fData.Width * y + x];
end;

procedure TgrayImage.setPixel(x, y: word; v: byte);
begin
 if (x < fData.Width) and (y < fData.Height)
 then fData.Data[fData.Width * y + x]:= v;
 fOneBit:= fOneBit and (v < 2);
end;

procedure TgrayImage.LoadFromStream(s: TStream);
begin
 LoadGrayImage(s, fData);
 DataChanged;
end;

procedure TgrayImage.SaveToStream(s: TStream);
begin
 SaveGrayImage(s, fData);
end;

function TgrayImage.getScanline(y: word): PByteArray;
begin
 if y >= fData.Height
 then raise Exception.Create('Wrong coordinates');
 Result:= PByteArray(@(fData.Data[fData.Width * y]));
end;

procedure TgrayImage.DataChanged;
var
 c: integer;
begin
 fOneBit:= True;
 for c:= 0 to integer(Width * Height) - 1 do if Data[c] > 1 then begin
  fOneBit:= False;
  Break;
 end;
end;

procedure TgrayImage.FindObjects(cbp: TfinderCallBackProc; const rect: TRect; const invert: boolean = False);
var
 x, y, x1, x2, y1, y2: word;
 c, cY, cYM, cYP: cardinal;
 a, d: byte;
 dt: TBytes;
 levels: array [byte] of cardinal;
 r: TRect;
 stop: boolean;

 function findObject(sX, sY: word): TRect;
 type
  tVector = (veUp, veDown, veLeft, veRight, veStop);
 const
  veArray: array [-1..1,-1..1] of tVector = (
   (veStop, veLeft, veStop),
   (veUp, veStop, veDown),
   (veStop, veRight, veStop));
  markValue = 3;
 var
  dY, dYM, dYP: cardinal;
  pX, pY,
  x, y, x1, x2, y1, y2: word;
  vector: tVector;
  started: boolean;
 begin
  x:= sX;
  y:= sY;
  x1:= x;
  x2:= x;
  y1:= y;
  y2:= y;
  vector:= veRight;
  repeat
   pX:= x;
   pY:= y;
   dY:= Width * y;
   dYM:= dY - Width;
   dYP:= dY + Width;

   case vector of
   veRight: begin
     if dt[dYM + x] > 0 then dec(y) else
     if dt[dY + x + 1] > 0 then inc(x) else
     if dt[dYP + x] > 0 then inc(y) else
     if dt[dY + x - 1] > 0 then dec(x)
    end;
   veDown: begin
     if dt[dY + x + 1] > 0 then inc(x) else
     if dt[dYP + x] > 0 then inc(y) else
     if dt[dY + x - 1] > 0 then dec(x) else
     if dt[dYM + x] > 0 then dec(y)
    end;
   veLeft: begin
     if dt[dYP + x] > 0 then inc(y) else
     if dt[dY + x - 1] > 0 then dec(x) else
     if dt[dYM + x] > 0 then dec(y) else
     if dt[dY + x + 1] > 0 then inc(x)
    end;
   veUp: begin
     if dt[dY + x - 1] > 0 then dec(x) else
     if dt[dYM + x] > 0 then dec(y) else
     if dt[dY + x + 1] > 0 then inc(x) else
     if dt[dYP + x] > 0 then inc(y)
    end;
   end;

   dt[dY + x]:= markValue; // used!
   vector:= veArray[x - pX, y - pY];
   if x1 > x then x1:= x;
   if x2 < x then x2:= x;
   if y1 > y then y1:= y;
   if y2 < y then y2:= y;
  until (x = sX) and (y = sY) or (vector = veStop);
  for y:= y1 to y2 - 1 do begin
   started:= False;
   dY:= Width * y;
   dYM:= dY - Width;
   dYP:= dY + Width;
   for x:= x1 + 1 to x2 do begin
    started:= (started or (dt[dYM + x] = markValue) or
        (dt[dY + x - 1] = markValue) or
        (dt[dYP + x] = markValue)) and (dt[dY + x] > 0);
    if started then dt[dY + x]:= markValue;
   end
  end;
  for y:= y2 downto y1 + 1 do begin
   started:= False;
   dY:= Width * y;
   dYM:= dY - Width;
   dYP:= dY + Width;
   for x:= x2 - 1 downto x1 do begin
    started:= (started or (dt[dYP + x] = markValue) or
        (dt[dY + x + 1] = markValue) or
        (dt[dYM + x] = markValue)) and (dt[dY + x] > 0);
    if started then dt[dY + x]:= markValue;
   end
  end;
  Result.Top:= y1;
  Result.Left:= x1;
  Result.Right:= x2 + 1;
  Result.Bottom:= y2 + 1;
 end;

begin
 if (Width < 3) or (Height < 3) then exit;
 if rect.Left < 1 then x1:= 1 else if rect.Left > width - 2 then x1:= Width - 1 else x1:= rect.Left;
 if rect.Top < 1 then y1:= 1 else if rect.Top > Height - 2 then y1:= Height - 1 else y1:= rect.Top;
 if rect.Right < 1 then x2:= 1 else if rect.Right > width - 3 then x2:= Width - 2 else x2:= rect.Right;
 if rect.Bottom < 1 then y2:= 1 else if rect.Bottom > Height - 3 then y2:= Height - 2 else y2:= rect.Bottom;
 setLength(dt, length(Data));
 if fOneBit then begin
  for y:= y1 to y2 do begin
   cY:= y * Width;
   for x:= x1 to x2 do begin
    c:= cY + x;
    dt[c]:= 0;
    if invert then dt[c]:= Data[c]
    else if Data[c] = 0 then dt[c]:= 1 else dt[c]:= 0;
   end;
  end;
 end else begin
  for c:= 0 to 255 do levels[c]:= 0;
  if invert then for y:= y1 to y2 do begin
   cY:= Width * y;
   for x:= x1 to x2 do begin
    c:= cY + x;
    cYM:= c - Width;
    cYP:= c + Width;
    a:= byte(word(Data[cYM] + Data[c - 1] + Data[c + 1] + Data[cYP] +
      Data[cYM - 1] + Data[cYP] + Data[cYP - 1] + Data[cYP + 1]
     ) shr 3);
    d:= Data[c];
    if a < d then d:= d - a else d:= 0;
    levels[d]:= levels[d] + 1;
    dt[c]:= d;
   end;
  end else for y:= y1 to y2 do begin
   cY:= Width * y;
   for x:= x1 to x2 do begin
    c:= cY + x;
    cYM:= c - Width;
    cYP:= c + Width;
    a:= byte(word(Data[cYM] + Data[c - 1] + Data[c + 1] + Data[cYP] +
      Data[cYM - 1] + Data[cYP] + Data[cYP - 1] + Data[cYP + 1]
     ) shr 3);
    d:= Data[c];
    if a > d then d:= a - d else d:= 0;
    levels[d]:= levels[d] + 1;
    dt[c]:= d;
   end;
  end;
  cY:= 0;
  c:= (x2 - x1) * (y2 - y1) shr 1;
  for d:= 255 downto 0 do begin
   inc(cY, levels[d]);
   if cY > c then break;
  end;
  a:= d shr 1;
  for y:= y1 to y2 do begin
   cY:= y * Width;
   for x:= x1 to x2 do begin
    c:= cY + x;
    if (dt[c] > d) or
       (dt[c] > a) and ((dt[c - 1] > 0) or (dt[c + 1] > a) or
       (dt[c - Width] > 0) or (dt[c + Width] > a))
    then dt[c]:= 1 else dt[c]:= 0;
   end;
  end;
 end;
 stop:= False;
 for y:= y1 to y2 do begin
  cY:= y * Width;
  for x:= x1 to x2 do begin
   c:= cY + x;
   if dt[c] = 1 then begin
    r:= findObject(x, y);
    if ((r.Right - r.Left) > 3) and ((r.Bottom - r.Top) > 3) then begin
     cbp(Self, r, stop);
     if stop then begin
      setLength(dt, 0);
      exit
     end;
    end;
   end;
  end;
 end;
 setLength(dt, 0);
end;

procedure TgrayImage.CopyRect(r: TRect; src: TgrayImage);
var
 x, y: cardinal;
begin
 setWidth(r.Width);
 setHeight(r.Height);
 for x:= r.Left to r.Right - 1 do
  for y:= r.Top to r.Bottom - 1 do
   Pixel[x - r.Left, y - r.Top]:= src.Pixel[x, y];
 DataChanged;
end;

procedure TgrayImage.Copy(src: TgrayImage);
begin
 Copy(src.rawData);
end;

procedure TgrayImage.Copy(src: TgrayImageData);
begin
 setWidth(src.Width);
 setHeight(src.Height);
 move(src.Data[0], Data[0], Width * Height);
 DataChanged;
end;

procedure TgrayImage.SetData(const w, h: word; const dt: PByte);
begin
 fData.Width:= w;
 fData.Height:= h;
 setLength(fData.Data, fData.Width * fData.Height);
 move(dt[0], fData.Data[0], w * h);
 DataChanged;
end;

procedure TgrayImage.Normalize;
var
 mn, mx: byte;
 k: extended;
 c: integer;
begin
 mn:= 255; mx:= 0;
 for c:= 0 to integer(Width * Height) - 1 do begin
  if Data[c] > mx then mx:= Data[c];
  if Data[c] < mn then mn:= Data[c];
 end;
 if (mn > 0) or (mx < 255) then begin
  k:= 255 / (mx - mn);
  for c:= 0 to integer(Width * Height) - 1 do Data[c]:= round((Data[c] - mn) * k);
  DataChanged;
 end;
end;

procedure TgrayImage.Convert1Bit(const invert: boolean = False);
const
 b: array [boolean, boolean] of byte = ((0, 1), (1, 0));
var
 z, mn, mx: word;
 a, c, x, y: integer;
begin
 mn:= 255; mx:= 0;
 for x:= 3 to Width - 4 do for y:= 3 to Height - 4 do begin
  z:= Pixel[x, y];
  if z > mx then mx:= z;
  if z < mn then mn:= z;
 end;
 a:= (mx + mn) shr 1;
 for c:= 0 to integer(Width * Height) - 1 do Data[c]:= b[invert, Data[c] > a];
 fOneBit:= True;
end;

procedure TgrayImage.Normalize1Bit(const sampleLength: word = 20);
var
 dt: TBytes;
 slx, sly: word;
 c: integer;
 a: word;

 function colorAround(x, y: word): byte;
 var
  i, i1, i2: integer;
  sy, s1, s2: cardinal;
 begin
  sy:= y * Width;
  s1:= 0;
  i1:= x - slx;
  if i1 < 0 then begin s1:= -i1; i1:= 0; end;
  i2:= x + slx + s1;
  if i2 >= Width then begin i1:= i1 - i2 + Width; i2:= Width - 1; end;
  s1:= 0;
  inc(i1, sy); inc(i2, sy);
  for i:= i1 to i2 do inc(s1, Data[i]);
  s1:= s1 div (slx shl 1);
  s2:= 0;
  i1:= y - sly;
  if i1 < 0 then begin s2:= -i1; i1:= 0; end;
  i2:= y + sly + s2;
  if i2 >= Height then begin i1:= i1 - i2 + Height; i2:= Height - 1; end;
  s2:= 0;
  sy:= x + i1 * Width;
  for i:= i1 to i2 do begin inc(s2, Data[sy]); inc(sy, Width); end;
  s2:= s2 div (sly shl 1);
  sy:= (s1 + s2) shr 1;
  if sy > 255 then sy:= 255;
  Result:= sy;
 end;

begin
 if fOneBit or (Width < 3) or (Height < 3) then exit;
 setLength(dt, length(Data));
 if sampleLength > Width then slx:= Width shr 1 else slx:= sampleLength shr 1;
 if sampleLength > Height then sly:= Height shr 1 else sly:= sampleLength shr 1;
 for c:= 0 to length(Data) - 1 do begin
  a:= colorAround(c mod Width, c div Width) * 15 shr 4;
  if Data[c] < a
  then dt[c]:= 0 else dt[c]:= 1;
 end;
 Move(dt[0], Data[0], length(dt));
 setLength(dt, 0);
 fOneBit:= True;
end;

//################################################################### TabImages

constructor TabImages.Create(aHeight: word; aAngle: integer; aRatio: byte);
begin
 fHeight:= aHeight;
 fAngle:= aAngle;
 fRatio:= aRatio;
end;

destructor TabImages.Destroy;
var
 c: integer;
begin
 for c:= 0 to Count - 1 do setLength(Images[c].image.Data, 0);
 setLength(Images, 0);
 inherited;
end;

procedure TabImages.LoadFromStream(s: TStream);
var
 i: integer;
 c: word;
begin
 s.Read(fHeight, SizeOf(fHeight));
 s.Read(fAngle, SizeOf(fAngle));
 s.Read(fRatio, SizeOf(fRatio));
 s.Read(c, SizeOf(c));
 setLength(Images, c);
 for i:= 0 to c - 1 do with Images[i] do begin
  s.Read(symbol, 1);
  s.Read(maxWeight, SizeOf(maxWeight));
  LoadGrayImage(s, image);
 end;
end;

procedure TabImages.SaveToStream(s: TStream);
var
 i: integer;
 c: word;
begin
 c:= Count;
 s.Write(fHeight, SizeOf(fHeight));
 s.Write(fAngle, SizeOf(fAngle));
 s.Write(fRatio, SizeOf(fRatio));
 s.Write(c, SizeOf(c));
 for i:= 0 to c - 1 do with Images[i] do begin
  s.Write(symbol, 1);
  s.Write(maxWeight, SizeOf(maxWeight));
  SaveGrayImage(s, image);
 end;
end;

function TabImages.getCount: integer;
begin
 Result:= length(Images);
end;

//################################################################### TbasicOCR

constructor TbasicOCR.Create;
begin
 minWeight:= 75;
 fText:= '';
 fCharWeight:= 0;
end;

destructor TbasicOCR.Destroy;
begin
 ClearSymbols;
 inherited;
end;

procedure TbasicOCR.ClearSymbols;
var
 c: integer;
begin
 for c:= 0 to length(Images) - 1 do Images[c].Free;
 setLength(Images, 0);
end;

procedure TbasicOCR.LoadFromStream(s: TStream);
var
 i: integer;
 c: word;
begin
 for i:= 0 to length(Images) - 1 do Images[i].Free;
 s.Read(c, SizeOf(c));
 setLength(Images, c);
 for i:= 0 to c - 1 do begin
  Images[i]:= TabImages.Create(0, 0, 100);
  Images[i].LoadFromStream(s);
 end;
end;

procedure TbasicOCR.LoadFromFile(const fn: string);
var
 f: TFileStream;
begin
 if FileExists(fn) then begin
  f:= TFileStream.Create(fn, fmOpenRead);
  try
   LoadFromStream(f);
  finally
   f.Free;
  end;
 end;
end;

function TbasicOCR.SearchChar(img: TgrayImage; r: TRect; h: byte; a: integer;
  p: byte; valid: ansiString): ansiChar;
var
 ci, cs, x, y, sx, sy, ax, ay, yw,
 tw, xw: integer;
begin
 Result:= #0;
 fCharWeight:= 0;
 if r.Top > 2 then ay:= r.Top - 3 else ay:= 0;
 if r.Left > 2 then ax:= r.Left - 3 else ax:= 0;
 for ci:= 0 to length(Images) - 1 do with Images[ci] do begin
  if (Height = h) and (Angle = a) and (Ratio = p) then begin
   for cs:= 0 to length(Images) - 1 do with Images[cs] do begin
    if (abs(image.Width - 2 - r.Width) < 3) and (pos(symbol, valid) > 0)
    then begin
     xw:= 0;
     for sx:= ax to r.Left + 1 do for sy:= ay to r.Top + 1 do begin
      tw:= 0;
      for y:= 0 to image.Height - 1 do begin
       yw:= y * image.Width;
       for x:= 0 to image.Width - 1
        do inc(tw, ocrWeights[img.Pixel[sx + x, sy + y], image.Data[x + yw]]);
      end;
      if tw > xw
      then xw:= tw;
     end;
     xw:= xw * 100 div maxWeight;
     if (xw > fCharWeight) and (xw > minWeight) then begin
      fCharWeight:= xw;
      Result:= symbol;
     end;
    end;
   end;
  end;
 end;
end;

function TbasicOCR.Search(const img: TgrayImage; const r: TRect): boolean;
begin
 fText:= '';
 Result:= False;
end;

function TbasicOCR.Search(const img: TgrayImage): boolean;
begin
 Result:= Search(img, rect(0, 0, img.Width, img.Height));
end;

end.

